package gov.usgs.earthquake.nshmp.netcdf.reader;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.netcdf.NshmNetcdfReader;
import gov.usgs.earthquake.nshmp.netcdf.reader.NetcdfUtils.Key;

import ucar.ma2.Array;
import ucar.ma2.DataType;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.dataset.NetcdfDataset;

/**
 * Container for gridded hazard curves at four closest grid points to target
 * 
 * @author U.S. Geological Survey
 */
public class BoundingHazardsReader {

  private final NshmNetcdfReader netcdf;
  private final NetcdfCoordinates coords;
  private BoundingHazards boundingHazards;
  private List<BoundingLocation> boundingLocations = new ArrayList<>();

  BoundingHazardsReader(NshmNetcdfReader netcdf, Location site) {
    this.netcdf = netcdf;
    this.coords = netcdf.coordinates();
    coords.contains(site);
    setBoundingHazards(site);
  }

  /**
   * Returns the bounding hazards at four closet grid points to target.
   *
   * @param netcdf The {@code Netcdf}
   * @param site The site to get bounding hazards
   */
  public static BoundingHazards boundingHazards(
      NshmNetcdfReader netcdf,
      Location site) {
    return new BoundingHazardsReader(netcdf, site).boundingHazards;
  }

  LocationList boundingLocations() {
    var locations = boundingLocations.stream()
        .map(boundingLocation -> boundingLocation.location)
        .collect(Collectors.toList());

    return LocationList.copyOf(locations);
  }

  /**
   * Get data for target point
   *
   * @param d1 data at first point (p1)
   * @param d2 data at second point (p2)
   * @param frac fractional distance between p1 and p2 to target point
   */
  static StaticHazards getTargetData(
      StaticHazards d1,
      StaticHazards d2,
      double frac) {
    NetcdfUtils.checkBoundingHazard(d1, d2);
    return frac == 0.0 ? d1 : frac == 1.0 ? d2 : NetcdfUtils.linearInterpolate(d1, d2, frac);
  }

  private void setBoundingHazards(Location site) {
    var longitudes = coords.longitudes();
    var latitudes = coords.latitudes();

    var idxLonLL = NetcdfUtils.getIdxLTEQ(longitudes, site.longitude);
    var idxLatLL = NetcdfUtils.getIdxLTEQ(latitudes, site.latitude);

    var lonLeft = longitudes[idxLonLL];
    var lonRight = longitudes[idxLonLL + 1];
    var latLower = latitudes[idxLatLL];
    var latUpper = latitudes[idxLatLL + 1];

    boundingLocations.add(new BoundingLocation(lonLeft, latLower, 0, 0));
    boundingLocations.add(new BoundingLocation(lonLeft, latUpper, 0, 1));
    boundingLocations.add(new BoundingLocation(lonRight, latUpper, 1, 1));
    boundingLocations.add(new BoundingLocation(lonRight, latLower, 1, 0));

    var hazards = extractHazardsAt(idxLonLL, idxLatLL);

    var fracLon = NetcdfUtils.calcGridFrac(longitudes, idxLonLL, site.longitude);
    var fracLat = NetcdfUtils.calcGridFrac(latitudes, idxLatLL, site.latitude);

    var builder = BoundingHazards.builder();
    hazards.forEach((key, value) -> builder.put(key, value));
    builder.put(
        site,
        calcTargetHazards(hazards, fracLon, fracLat))
        .build();
    boundingHazards = builder.build();

    NetcdfUtils.checkBoundingHazards(boundingHazards, boundingLocations.get(0).location);
  }

  private StaticHazards calcTargetHazards(BoundingHazards hazards, double fracLon, double fracLat) {
    var westTarget = getTargetData(
        hazards.get(boundingLocations.get(0).location),
        hazards.get(boundingLocations.get(1).location),
        fracLat);

    var eastTarget = getTargetData(
        hazards.get(boundingLocations.get(3).location),
        hazards.get(boundingLocations.get(2).location),
        fracLat);

    return getTargetData(westTarget, eastTarget, fracLon);
  }

  /*
   * Return hazard curves for four closest grid points by SiteClass and Imt
   */
  private BoundingHazards extractHazardsAt(
      int idxLonLL,
      int idxLatLL) {
    try (NetcdfDataset ncd = NetcdfDataset.openDataset(netcdf.path().toString())) {
      var boundingHazardMaps = BoundingHazards.builder();
      var targetGroup = ncd.findGroup(netcdf.nshmGroup().baseGroup());

      var targetOrigin = new int[] { 0, 0, idxLatLL, idxLonLL, 0 };
      var targetShape = new int[] {
          coords.siteClasses().size(),
          coords.imts().size(),
          2,
          2,
          coords.nIml()
      };

      // TODO: rename variable in netCDF
      /*
       * Array aHazards now has shape [nVs,nImt,2,2,nIml] ...so origin will now
       * be [0,0,0,0,0] for LL grid point ...and shape of requested array is
       * [nVs,nImt,1,1,nIml]
       */
      var aHazards = targetGroup.findVariable(Key.AEPS).read(targetOrigin, targetShape);

      var shape = targetShape.clone();
      shape[2] = 1;
      shape[3] = 1;

      for (var boundingLocation : boundingLocations) {
        boundingHazardMaps.put(
            boundingLocation.location,
            mapHazardsFromArray(aHazards.section(boundingLocation.origin, shape)));
      }

      return boundingHazardMaps.build();
    } catch (IOException | InvalidRangeException e) {
      throw new RuntimeException("Could not read Netcdf file [" + netcdf.path() + "]");
    }
  }

  /*
   * Read hazard curves from netCDF variable into map of hazards by SiteClass
   * and Imt
   *
   * TODO: if target is on a grid point (or on a grid lat or lon), no need to
   * read 4 bounding points ?
   */
  private StaticHazards mapHazardsFromArray(Array hazards) {
    var vsImtHazardMap = StaticHazards.builder();

    for (int iSiteClass = 0; iSiteClass < coords.siteClasses().size(); iSiteClass++) {
      var siteClass = coords.siteClasses().get(iSiteClass);

      var imtHazardMap = StaticHazard.builder();
      for (int iImt = 0; iImt < coords.imts().size(); iImt++) {
        var imt = coords.imts().get(iImt);

        var origin = new int[] { iSiteClass, iImt, 0 };
        var shape = new int[] { 1, 1, coords.nIml() };

        try {
          var xySequence = XySequence.create(
              coords.imls().get(siteClass).get(imt),
              (double[]) hazards.section(origin, shape).reduce().get1DJavaArray(DataType.DOUBLE));

          imtHazardMap.put(imt, xySequence);
        } catch (InvalidRangeException e) {
          throw new RuntimeException(e.getMessage());
        }
      }

      vsImtHazardMap.put(siteClass, imtHazardMap.build());
    }

    return vsImtHazardMap.build();
  }

  static class BoundingLocation {
    final Location location;
    final int[] origin;

    BoundingLocation(double longitude, double latitude, int longitudeIndex, int latitudeIndex) {
      location = Location.create(latitude, longitude);
      origin = new int[] { 0, 0, latitudeIndex, longitudeIndex, 0 };
    }
  }

}
