package gov.usgs.earthquake.nshmp.netcdf.www;

import java.nio.file.Path;

import javax.annotation.Nullable;
import javax.inject.Inject;

import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.internal.www.NshmpMicronautServlet;
import gov.usgs.earthquake.nshmp.netcdf.NshmGroup;
import gov.usgs.earthquake.nshmp.netcdf.SiteClass;

import io.micronaut.context.annotation.Value;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.swagger.v3.oas.annotations.Hidden;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;

/**
 * Micronaut controller for getting static hazards for a specific
 * {@code NshmGroup} from a NetCDF file.
 * 
 * @see NetcdfService
 * 
 * @author U.S. Geological Survey
 */
@Tag(name = "Hazard Data")
@Controller("/data")
public class NetcdfController {

  @Inject
  private NshmpMicronautServlet servlet;

  @Value("${nshmp-netcdf.nshm-group}")
  NshmGroup nshmGroup;

  @Value("${nshmp-netcdf.netcdf-file}")
  Path path;

  /**
   * GET method to return the usage information on the NetCDF web service.
   * 
   * @param request The HTTP request
   */
  @Operation(
      summary = "Returns the web service usage information with current NSHM",
      description = "Returns the supported:\n * longitudes\n * latitudes\n * site classes\n * IMTs",
      operationId = "netcdf_data_doGetMetadata")
  @ApiResponse(
      description = "NSHM NetCDF usage",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/usage", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetMetadata(HttpRequest<?> request) {
    return NetcdfService.metadata(nshmGroup, path, servlet.urlHelper(request));
  }

  /**
   * GET method to return a hazard curve using URL query.
   * 
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class (optional)
   * @param imt The IMT (optional)
   */
  @Operation(
      summary = "Returns hazard curve(s) given a longitude, latitude, and/or a site class, and imt",
      description = "Retrieve hazard curve(s) from a NSHM NetCDF file.\n\n" +
          "For supported longitudes, latitudes, site classes, and IMTs see the usage information.",
      operationId = "netcdf_data_doGetHazard")
  @ApiResponse(
      description = "Returns a hazard curve from the NSHM NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "{?longitude,latitude,siteClass,imt}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetHazard(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @Schema(required = true) @QueryValue @Nullable Double latitude,
      @QueryValue @Nullable SiteClass siteClass,
      @QueryValue @Nullable Imt imt) {
    var urlHelper = servlet.urlHelper(request);
    var query = new Query(longitude, latitude, siteClass, imt);
    return NetcdfService.handleDoGet(nshmGroup, path, urlHelper, query);
  }

  /**
   * GET method to return a hazard curve using slash delimited.
   * 
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class
   * @param imt The IMT
   */
  @Operation(
      summary = "Returns a hazard curve given a longitude, latitude, site class, and imt",
      description = "Retrieve a hazard curve from a NSHM NetCDF file.\n\n" +
          "For supported longitudes, latitudes, site classes, and IMTs see the usage information.",
      operationId = "netcdf_data_doGetHazardSlashWithSiteClassAndImt")
  @ApiResponse(
      description = "Returns a hazard curve from the NSHM NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{longitude}/{latitude}/{siteClass}/{imt}")
  public HttpResponse<String> doGetHazardSlashBySiteImt(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable Double latitude,
      @Schema(required = true) @PathVariable @Nullable SiteClass siteClass,
      @Schema(required = true) @PathVariable @Nullable Imt imt) {
    return doGetHazard(request, longitude, latitude, siteClass, imt);
  }

  /**
   * GET method to return hazard curves using slash delimited.
   * 
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class
   */
  @Operation(
      summary = "Returns hazard curves given a longitude, latitude, and site class.",
      description = "Retrieve hazard curves from a NSHM NetCDF file.\n\n" +
          "For supported longitudes, latitudes, and site classes see the usage information.",
      operationId = "netcdf_data_doGetHazardSlashWithSiteClass")
  @ApiResponse(
      description = "Returns hazard curves from the NSHM NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{longitude}/{latitude}/{siteClass}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetHazardSlashBySite(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable Double latitude,
      @Schema(required = true) @PathVariable @Nullable SiteClass siteClass) {
    return doGetHazard(request, longitude, latitude, siteClass, null);
  }

  /**
   * GET method to return hazard curves using slash delimited.
   * 
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   */
  @Operation(
      summary = "Returns hazard curves given a longitude and latitude.",
      description = "Retrieve hazard curves from a NSHM NetCDF file.\n\n" +
          "For supported longitudes and latitudes see the usage information.",
      operationId = "netcdf_data_doGetHazardSlash")
  @ApiResponse(
      description = "Returns hazard curves from the NSHM NetCDF file",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{longitude}/{latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetHazardSlash(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable @Nullable Double longitude,
      @Schema(required = true) @PathVariable @Nullable Double latitude) {
    return doGetHazard(request, longitude, latitude, null, null);
  }

  /**
   * Returns the border as a GeoJSON feature collections.
   * 
   * @param request The HTTP request
   */
  @Operation(
      summary = "Returns the border as a GeoJSON feature collection.",
      description = "Returns a GeoJSON feature collections of the border defining the data.",
      operationId = "netcdf_data_region_doGetBorder")
  @ApiResponse(
      description = "GeoJSON feature collection",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/border", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetBorder(HttpRequest<?> request) {
    var urlHelper = servlet.urlHelper(request);
    return NetcdfService.handleDoGetBorder(nshmGroup, path, urlHelper);
  }

  /**
   * GET method to return the bounding hazards using URL query.
   * 
   * @param request The HTTP request
   * @param longitude The longitude
   * @param latitude The latitude
   */
  @Hidden
  @Get(uri = "/bounding{?longitude,latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetHazardBounding(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @Schema(required = true) @QueryValue @Nullable Double latitude) {
    var urlHelper = servlet.urlHelper(request);
    var query = new Query(longitude, latitude, null, null);
    return NetcdfService.handleDoGetBounding(nshmGroup, path, urlHelper, query);
  }

  /**
   * GET method to return the bounding hazards using slash delimited.
   * 
   * @param request The HTTP request
   * @param longitude The longitude
   * @param latitude The latitude
   */
  @Hidden
  @Get(uri = "/bounding/{longitude}/{latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetHazardBoundingSlash(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @Schema(required = true) @QueryValue @Nullable Double latitude) {
    return doGetHazardBounding(request, longitude, latitude);
  }

  static class Query {
    final Double longitude;
    final Double latitude;
    final SiteClass siteClass;
    final Imt imt;

    Query(Double longitude, Double latitude, SiteClass siteClass, Imt imt) {
      this.longitude = longitude;
      this.latitude = latitude;
      this.siteClass = siteClass;
      this.imt = imt;
    }
  }

  static enum Service {
    BOUNDING,
    CURVES,
    CURVES_BY_SITE_CLASS,
    CURVES_BY_SITE_CLASS_IMT;
  }

}
