package gov.usgs.earthquake.nshmp.netcdf;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import com.google.common.io.Resources;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.netcdf.reader.BoundingHazards;
import gov.usgs.earthquake.nshmp.netcdf.reader.StaticHazard;
import gov.usgs.earthquake.nshmp.netcdf.reader.StaticHazards;

public class NshmNetcdfReaderTests {

  static final String CONUS_TEST_FILE = "conus-test.nc";
  static final Path NETCDF_PATH = Paths.get(Resources.getResource(CONUS_TEST_FILE).getPath());

  static final String TEST_INVALID_NC_FILE = "invalid-netcdf-file.nc";
  static final Path NETCDF_INVALID_FILE = Paths.get(
      Resources.getResource(TEST_INVALID_NC_FILE).getPath());

  // difference tolerance, until we can incorporate precision into
  // NshmNetcdfReader
  public static final double IML_TOL = 1e-6;
  static final double HAZ_TOL = 1e-8;

  static final Location TARGET_LOCATION = Location.create(39.213, -105.234);

  public static final double[] EXPECTED_LONGITUDES =
      new double[] { -105.3, -105.25, -105.2, -105.15 };
  public static final double[] EXPECTED_LATITUDES = new double[] { 39.15, 39.2, 39.25, 39.3 };

  static final LocationList BOUNDING_LOCATIONS = LocationList.builder()
      .add(EXPECTED_LATITUDES[1], EXPECTED_LONGITUDES[1])
      .add(EXPECTED_LATITUDES[2], EXPECTED_LONGITUDES[1])
      .add(EXPECTED_LATITUDES[2], EXPECTED_LONGITUDES[2])
      .add(EXPECTED_LATITUDES[1], EXPECTED_LONGITUDES[2])
      .build();

  static final LocationList LOCATIONS = LocationList.builder()
      .addAll(BOUNDING_LOCATIONS)
      .add(TARGET_LOCATION)
      .build();

  static final List<double[]> XS = List.of(
      new double[] { 0.00233, 0.0265, 0.302, 3.44 },
      new double[] { 0.00333, 0.0379, 0.432, 4.92 });

  static final List<double[]> HAZARDS = List.of(
      new double[] { 0.04387208, 0.004250093, 0.0001192809, 2.201278e-07 },
      new double[] { 0.06567874, 0.005199003, 6.54093e-05, 6.705638e-08 },
      new double[] { 0.04127881, 0.003905505, 0.000122145, 3.597946e-07 },
      new double[] { 0.05262404, 0.003470381, 4.079298e-05, 4.619109e-08 },
      new double[] { 0.04388691, 0.004249376, 0.0001185702, 2.198896e-07 },
      new double[] { 0.06564134, 0.005175665, 6.480198e-05, 6.684925e-08 },
      new double[] { 0.04130831, 0.003907337, 0.0001214354, 3.59386e-07 },
      new double[] { 0.05258384, 0.003454936, 4.047655e-05, 4.606202e-08 },
      new double[] { 0.04346677, 0.00412513, 0.00011355, 2.171376e-07 },
      new double[] { 0.06530787, 0.0050781, 6.258566e-05, 6.466215e-08 },
      new double[] { 0.04089805, 0.003789999, 0.0001163111, 3.544692e-07 },
      new double[] { 0.0522491, 0.003379781, 3.912059e-05, 4.462088e-08 },
      new double[] { 0.04344459, 0.004124759, 0.0001141019, 2.173907e-07 },
      new double[] { 0.06532912, 0.005098603, 6.312435e-05, 6.484549e-08 },
      new double[] { 0.04086227, 0.003787166, 0.0001168667, 3.549025e-07 },
      new double[] { 0.05227672, 0.003393233, 3.939733e-05, 4.473717e-08 },
      new double[] { 4.3739751E-02, 4.2098902E-03, 1.1745205E-04, 2.1918876E-07 },
      new double[] { 6.5558481E-02, 5.1610430E-03, 6.4525923E-05, 6.6297021E-08 },
      new double[] { 4.1153710E-02, 3.8681961E-03, 1.2028426E-04, 3.5812084E-07 },
      new double[] { 5.2503492E-02, 3.4418438E-03, 4.0267402E-05, 4.5693341E-08 });

  public static final List<SiteClass> SITE_CLASSES = List.of(SiteClass.CD, SiteClass.C);
  public static final List<Imt> IMTS = List.of(Imt.PGA, Imt.SA0P4);

  static final int TARGET_LOWER_LEFT_LONGITUDE_IDX = 1;
  static final int TARGET_LOWER_LEFT_LATITUDE_IDX = 1;

  static final double TARGET_LONGITUDE_FRAC;
  static final double TARGET_LATITUDE_FRAC;

  public static Map<SiteClass, Map<Imt, double[]>> IMLS = new HashMap<>();

  static BoundingHazards BOUNDING_HAZARDS;

  public static final NshmNetcdfReader NETCDF =
      new NshmNetcdfReader(NshmGroup.CONUS_2018, NETCDF_PATH);

  static {
    var builder = BoundingHazards.builder();

    var iHaz = 0;
    for (var location : LOCATIONS) {
      var siteClassMap = StaticHazards.builder();

      for (var siteClass : SITE_CLASSES) {
        var imtMap = StaticHazard.builder();

        for (var iImt = 0; iImt < IMTS.size(); iImt++) {
          var imt = IMTS.get(iImt);
          var xy = XySequence.create(XS.get(iImt), HAZARDS.get(iHaz++));
          imtMap.put(imt, xy);
        }

        siteClassMap.put(siteClass, imtMap.build());
      }

      builder.put(location, siteClassMap.build());
    }

    BOUNDING_HAZARDS = builder.build();

    SITE_CLASSES.forEach(siteClass -> {
      var haz = new HashMap<Imt, double[]>();

      for (var i = 0; i < IMTS.size(); i++) {
        haz.put(IMTS.get(i), XS.get(i));
      }

      IMLS.put(siteClass, haz);
    });

    // calculate interpolation fraction for longitude and latitude
    int i = TARGET_LOWER_LEFT_LONGITUDE_IDX;
    var locationA = LOCATIONS.get(i);
    var locationB = LOCATIONS.get(i + 1);
    TARGET_LONGITUDE_FRAC = (TARGET_LOCATION.longitude - locationA.longitude) /
        (locationB.longitude - locationA.longitude);

    i = TARGET_LOWER_LEFT_LATITUDE_IDX;
    TARGET_LATITUDE_FRAC = (TARGET_LOCATION.latitude - locationA.latitude) /
        (locationB.latitude - locationA.latitude);
  }

  @Test
  final void nshmGroupTests() {
    assertEquals(NshmGroup.CONUS_2018, NETCDF.nshmGroup());
    assertTrue(NshmGroup.CONUS_2018.equals(NETCDF.nshmGroup()));
  }

  @Test
  final void pathTests() {
    assertEquals(NETCDF_PATH, NETCDF.path());
    assertTrue(NETCDF_PATH.equals(NETCDF.path()));
  }

  @Test
  final void boundingHazardsTests() {
    var boundingHazards = NETCDF.boundingHazards(TARGET_LOCATION);

    assertEquals(BOUNDING_HAZARDS.size(), boundingHazards.size());

    for (var location : LOCATIONS) {
      assertTrue(boundingHazards.containsKey(location));
      assertTrue(BOUNDING_HAZARDS.containsKey(location));

      var expected = BOUNDING_HAZARDS.get(location);
      var actual = boundingHazards.get(location);
      testHazards(expected, actual);
    }
  }

  @ParameterizedTest(name = "{index} ==> Site: {0}")
  @MethodSource("byLocation")
  final void hazardTests(Location site) {
    var expected = BOUNDING_HAZARDS.get(site);
    var actual = NETCDF.hazard(site);
    testHazards(expected, actual);
  }

  @ParameterizedTest(name = "{index} ==> Site: {0}, {1}")
  @MethodSource("bySiteClass")
  final void hazardWithSiteClassTests(Location site, SiteClass siteClass) {
    var expected = BOUNDING_HAZARDS.get(site).get(siteClass);
    var actual = NETCDF.hazard(site, siteClass);
    testHazard(expected, actual);
  }

  @ParameterizedTest(name = "{index} ==> {0}, {1}, {2}")
  @MethodSource("bySiteClassImt")
  final void hazardWithSiteClassImtTests(Location site, SiteClass siteClass, Imt imt) {
    var expected = BOUNDING_HAZARDS.get(site).get(siteClass).get(imt);
    var actual = NETCDF.hazard(site, siteClass, imt);
    testSequence(expected, actual);
  }

  @Test
  final void invalidNetcdfFileTest() {
    assertThrows(IllegalArgumentException.class, () -> {
      new NshmNetcdfReader(NshmGroup.CONUS_2018, Paths.get("fileDoesNotExist"));
    });

    assertThrows(RuntimeException.class, () -> {
      new NshmNetcdfReader(NshmGroup.CONUS_2018, NETCDF_INVALID_FILE);
    });
  }

  private static Stream<Location> byLocation() {
    return LOCATIONS.stream();
  }

  private static Stream<? extends Arguments> bySiteClass() {
    return byLocation().flatMap(location -> {
      return SITE_CLASSES.stream().map(siteClass -> new Object[] { location, siteClass });
    }).map(Arguments::of);
  }

  private static Stream<? extends Arguments> bySiteClassImt() {
    return bySiteClass().map(args -> args.get())
        .flatMap(obj -> IMTS.stream().map(imt -> new Object[] { obj[0], obj[1], imt }))
        .map(Arguments::of);
  }

  private void testHazards(
      StaticHazards expected,
      StaticHazards actual) {
    for (var siteEntry : expected.entrySet()) {
      var siteClass = siteEntry.getKey();
      assertTrue(actual.containsKey(siteClass));
      assertTrue(expected.containsKey(siteClass));
      testHazard(expected.get(siteClass), actual.get(siteClass));
    }
  }

  private void testHazard(StaticHazard expected, StaticHazard actual) {
    for (var imtEntry : expected.entrySet()) {
      var imt = imtEntry.getKey();
      assertTrue(actual.containsKey(imt));
      assertTrue(expected.containsKey(imt));
      testSequence(expected.get(imt), actual.get(imt));
    }
  }

  private void testSequence(XySequence expected, XySequence actual) {
    assertArrayEquals(
        expected.xValues().toArray(),
        actual.xValues().toArray(),
        IML_TOL);

    assertArrayEquals(
        expected.yValues().toArray(),
        actual.yValues().toArray(),
        HAZ_TOL);
  }

}
